"""The ``input`` module allows you to access the conversion routines programmatically

To use the routines, you need to instanciate a ``ReadStrategy`` and a ``WriteStrategy``
relevant for the format of the input and output data.  You then pass these to a
``Context``.

Example
-------
Convert an in-memory dictionary of pandas DataFrames containing OSeMOSYS parameters
to an Excel spreadsheet::

>>> from otoole import ReadMemory
>>> from otoole import WriteExcel
>>> from otoole import Context
>>> reader = ReadMemory(parameters)
>>> writer = WriteExcel()
>>> converter = Context(read_strategy=reader, write_strategy=writer)
>>> converter.convert('.', 'osemosys_to_excel.xlsx')

Convert a GNUMathProg datafile to a folder of CSV files::

>>> from otoole import ReadDataFile
>>> from otoole import WriteCsv
>>> from otoole import Context
>>> reader = ReadDataFile()
>>> writer = WriteCsv()
>>> converter = Context(read_strategy=reader, write_strategy=writer)
>>> converter.convert('my_datafile.txt', 'folder_of_csv_files')

Convert a GNUMathProg datafile to a folder of Tabular DataPackage::

>>> from otoole import ReadDataFile
>>> from otoole import WriteDatapackage
>>> from otoole import Context
>>> reader = ReadDataFile()
>>> writer = WriteDatapackage()
>>> converter = Context(read_strategy=reader, write_strategy=writer)
>>> converter.convert('my_datafile.txt', 'my_datapackage')

"""
from __future__ import annotations

import logging
from abc import ABC, abstractmethod
from typing import Any, Dict, Optional, TextIO, Tuple, Union

import pandas as pd

from otoole.utils import read_packaged_file

logger = logging.getLogger(__name__)


class Inputs(object):
    """Represents the set of inputs associated with an OSeMOSYS model

    Arguments
    ---------
    config : str, default=None
    """

    def __init__(self, config: str = None):
        self.config = read_packaged_file("config.yaml", "otoole.preprocess")
        self._default_values = self._read_default_values()

    def _read_default_values(self):
        default_values = {}
        for name, contents in self.config.items():
            if contents["type"] == "param":
                default_values[name] = contents["default"]
        return default_values

    def default_values(self):
        return self._default_values


class Context:
    """
    The Context defines the interface of interest to clients.
    """

    def __init__(
        self, read_strategy: ReadStrategy, write_strategy: WriteStrategy
    ) -> None:
        """
        Usually, the Context accepts a strategy through the constructor, but
        also provides a setter to change it at runtime.
        """
        self._read_strategy = read_strategy
        self._write_strategy = write_strategy

    @property
    def write_strategy(self) -> WriteStrategy:
        """
        The Context maintains a reference to one of the Strategy objects. The
        Context does not know the concrete class of a strategy. It should work
        with all strategies via the Strategy interface.
        """
        return self._write_strategy

    @write_strategy.setter
    def write_strategy(self, strategy: WriteStrategy) -> None:
        """
        Usually, the Context allows replacing a Strategy object at runtime.
        """
        self._write_strategy = strategy

    @property
    def read_strategy(self) -> ReadStrategy:
        """
        The Context maintains a reference to one of the Strategy objects. The
        Context does not know the concrete class of a strategy. It should work
        with all strategies via the Strategy interface.
        """
        return self._read_strategy

    @read_strategy.setter
    def read_strategy(self, strategy: ReadStrategy) -> None:
        """
        Usually, the Context allows replacing a Strategy object at runtime.
        """
        self._read_strategy = strategy

    def _read(
        self, filepath: str, **kwargs
    ) -> Tuple[Dict[str, pd.DataFrame], Dict[str, Any]]:
        """Delegate reading to the strategy, depending upon the format
        """
        return self._read_strategy.read(filepath, **kwargs)

    def _write(self, inputs: Dict, filepath: str, default_values: Dict) -> None:
        """
        Delegate writing to the strategy, depending upon the format
        """
        self._write_strategy.write(inputs, filepath, default_values)

    def convert(self, input_filepath: str, output_filepath: str, **kwargs: Dict):
        """Converts from file ``input_filepath`` to file ``output_filepath``

        Arguments
        ---------
        input_filepath: str
        output_filepath: str
        """
        inputs, default_values = self._read(input_filepath, **kwargs)
        self._write(inputs, output_filepath, default_values)


class Strategy(ABC):
    """

    Arguments
    ---------
    input_config : dict, default=None
        A user configuration for the input parameters and sets
    results_config : dict, default=None
        A user configuration for the results parameters

    """

    def __init__(
        self, user_config: Optional[Dict] = None, results_config: Optional[Dict] = None
    ):
        self._input_config = {}
        self._results_config = {}

        if user_config:
            self._input_config = user_config
        else:
            self._input_config = self._read_config()
        if results_config:
            self._results_config = results_config
        else:
            self._results_config = self._read_results_config()

    def _read_config(self) -> Dict[str, Dict]:
        return read_packaged_file("config.yaml", "otoole.preprocess")

    def _read_results_config(self) -> Dict[str, Dict]:
        return read_packaged_file("config.yaml", "otoole.results")

    @property
    def input_config(self):
        return self._input_config

    @property
    def results_config(self):
        return self._results_config

    @staticmethod
    def _read_default_values(config):
        default_values = {}
        for name, contents in config.items():
            if contents["type"] == "param":
                default_values[name] = contents["default"]
        return default_values


class WriteStrategy(Strategy):
    """
    The WriteStrategy interface declares operations common to all writing formats

    The Context uses this interface to call the algorithm defined by Concrete
    Strategies.

    Arguments
    ---------
    filepath: str, default=None
    default_values: dict, default=None
    user_config: dict, default=None

    """

    def __init__(
        self,
        filepath: Optional[str] = None,
        default_values: Optional[Dict] = None,
        user_config: Optional[Dict] = None,
    ):
        super().__init__(user_config=user_config)
        if filepath:
            self.filepath = filepath
        else:
            self.filepath = ""

        if default_values:
            self.default_values = default_values
        else:
            self.default_values = {}

    @abstractmethod
    def _header(self) -> Union[TextIO, Any]:
        raise NotImplementedError()

    @abstractmethod
    def _write_parameter(
        self, df: pd.DataFrame, parameter_name: str, handle: TextIO, default: float
    ) -> pd.DataFrame:
        """Write parameter data"""
        raise NotImplementedError()

    @abstractmethod
    def _write_set(self, df: pd.DataFrame, set_name, handle: TextIO) -> pd.DataFrame:
        """Write set data"""
        raise NotImplementedError()

    @abstractmethod
    def _footer(self, handle: TextIO):
        raise NotImplementedError()

    def write(self, inputs: Dict, filepath: str, default_values: Dict):
        """Perform the conversion from dict of dataframes to destination format
        """
        self.filepath = filepath
        self.default_values = default_values

        handle = self._header()
        logger.debug(default_values)

        for name, df in sorted(inputs.items()):
            logger.debug(name)

            df = df.reset_index()
            if "index" in df.columns:
                df = df.drop(columns="index")

            logger.debug("Number of columns: %s, %s", len(df.columns), df.columns)
            if len(df.columns) > 1:
                default_value = default_values[name]
                self._write_parameter(df, name, handle, default=default_value)
            else:
                self._write_set(df, name, handle)

        self._footer(handle)

        if handle:
            handle.close()


class ReadStrategy(Strategy):
    """
    The Strategy interface declares operations common to all reading formats.

    The Context uses this interface to call the algorithm defined by Concrete
    Strategies.
    """

    def _check_index(self, input_data: Dict):
        """Checks and applied that an index is applied to the parameter DataFrame
        """
        for name, df in input_data.items():
            details = self.input_config[name]
            try:
                df.set_index(details["indices"], inplace=True)
            except KeyError:
                logger.debug("Parameter %s is indexed", name)

    @abstractmethod
    def read(
        self, filepath: Union[str, TextIO], **kwargs
    ) -> Tuple[Dict[str, pd.DataFrame], Dict[str, Any]]:
        raise NotImplementedError()



# -*- coding: utf-8 -*-
"""
    Dummy conftest.py for otoole.

    If you don't know what this is for, just leave it empty.
    Read more about conftest.py under:
    https://pytest.org/latest/plugins.html
"""

from pytest import fixture

import pandas as pd


@fixture
def annual_technology_emissions_by_mode():
    df = pd.DataFrame(
        data=[["SIMPLICITY", "GAS_EXTRACTION", "CO2", 1, 2014, 1.0]],
        columns=[
            "REGION",
            "TECHNOLOGY",
            "EMISSION",
            "MODE_OF_OPERATION",
            "YEAR",
            "VALUE",
        ],
    ).set_index(["REGION", "TECHNOLOGY", "EMISSION", "MODE_OF_OPERATION", "YEAR"])
    return df


@fixture
def discount_rate():
    df = pd.DataFrame(
        data=[["SIMPLICITY", "GAS_EXTRACTION", 0.05]],
        columns=["REGION", "TECHNOLOGY", "VALUE"],
    ).set_index(["REGION", "TECHNOLOGY"])

    return df


@fixture
def emission_activity_ratio():
    df = pd.DataFrame(
        data=[["SIMPLICITY", "GAS_EXTRACTION", "CO2", 1, 2014, 1.0]],
        columns=[
            "REGION",
            "TECHNOLOGY",
            "EMISSION",
            "MODE_OF_OPERATION",
            "YEAR",
            "VALUE",
        ],
    ).set_index(["REGION", "TECHNOLOGY", "EMISSION", "MODE_OF_OPERATION", "YEAR"])
    return df


@fixture
def emission_activity_ratio_two_techs():
    df = pd.DataFrame(
        data=[
            ["SIMPLICITY", "GAS_EXTRACTION", "CO2", 1, 2014, 1.0],
            ["SIMPLICITY", "DUMMY", "CO2", 1, 2014, 0.0],
        ],
        columns=[
            "REGION",
            "TECHNOLOGY",
            "EMISSION",
            "MODE_OF_OPERATION",
            "YEAR",
            "VALUE",
        ],
    ).set_index(["REGION", "TECHNOLOGY", "EMISSION", "MODE_OF_OPERATION", "YEAR"])

    return df


@fixture
def emissions_penalty():
    df = pd.DataFrame(
        data=[
            ["SIMPLICITY", "CO2", 2014, 1.23],
            ["SIMPLICITY", "CO2", 2015, 1.23],
            ["SIMPLICITY", "CO2", 2016, 1.23],
        ],
        columns=["REGION", "EMISSION", "YEAR", "VALUE"],
    ).set_index(["REGION", "EMISSION", "YEAR"])
    return df


@fixture
def yearsplit():
    df = pd.DataFrame(
        data=[
            ["ID", 2014, 0.1667],
            ["IN", 2014, 0.0833],
            ["SD", 2014, 0.1667],
            ["SN", 2014, 0.0833],
            ["WD", 2014, 0.3333],
            ["WN", 2014, 0.1667],
        ],
        columns=["TIMESLICE", "YEAR", "VALUE"],
    ).set_index(["TIMESLICE", "YEAR"])
    return df



"""Provides a command line interface to the ``otoole`` package

The key functions are **convert**, **cplex** and **viz**.

The ``convert`` command allows convertion of multiple different OSeMOSYS input formats
including from/to csv, an AMPL format datafile, a Tabular Data Package, a folder of
CSVs, an Excel workbook with one tab per parameter, an SQLite database

The ``cplex`` command provides access to scripts which transform and process a CPLEX
solution file into a format which is more readily processed - either to CBC or CSV
format.

The ``validate`` command checks the technology and fuel names of a tabular data package
against a standard or user defined configuration file.

The **viz** command allows you to produce a Reference Energy System diagram from a
Tabular Data Package.

Example
-------

Ask for help on the command line::

    >>> $ otoole --help
    usage: otoole [-h] [--verbose] [--version] {convert,cplex,validate,viz} ...

    otoole: Python toolkit of OSeMOSYS users

    positional arguments:
    {convert,cplex,validate,viz}
        convert             Convert from one input format to another
        cplex               Process a CPLEX solution file
        validate            Validate an OSeMOSYS model
        viz                 Visualise the model

    optional arguments:
    -h, --help            show this help message and exit
    --verbose, -v         Enable debug mode
    --version, -V         The version of otoole

"""
import argparse
import logging
import sys

from otoole import (
    ReadCbc,
    ReadCplex,
    ReadCsv,
    ReadDatafile,
    ReadDatapackage,
    ReadExcel,
    WriteCsv,
    WriteDatafile,
    WriteDatapackage,
    WriteExcel,
    __version__,
)
from otoole.input import Context
from otoole.utils import read_packaged_file
from otoole.validate import main as validate
from otoole.visualise import create_res

logger = logging.getLogger(__name__)


def validate_model(args):
    file_format = args.format

    if args.config:
        config = read_packaged_file(args.config)
        validate(file_format, args.filepath, config)
    else:
        validate(file_format, args.filepath)


def cplex2cbc(args):
    ReadCplex()._convert_cplex_file(
        args.cplex_file,
        args.output_file,
        args.start_year,
        args.end_year,
        args.output_format,
    )


def result_matrix(args):
    """Post-process results from CBC solution file into CSV format

    """
    msg = "Conversion from {} to {} is not yet implemented".format(
        args.from_format, args.to_format
    )

    read_strategy = None
    write_strategy = None

    if args.from_format == "cbc":
        read_strategy = ReadCbc()
    elif args.from_format == "cplex":
        read_strategy = ReadCplex()

    if args.to_format == "csv":
        write_strategy = WriteCsv()

    if args.input_datapackage:
        input_data, _ = ReadDatapackage().read(args.input_datapackage)
    elif args.input_datafile:
        input_data, _ = ReadDatafile().read(args.input_datafile)
    else:
        input_data = {}

    if read_strategy and write_strategy:
        context = Context(read_strategy, write_strategy)
        context.convert(args.from_path, args.to_path, input_data=input_data)
    else:
        raise NotImplementedError(msg)


def conversion_matrix(args):
    """Convert from one format to another

    Implemented conversion functions::

        from\to     ex cs dp df
        -----------------------
        excel       -- yy -- --
        csv         nn -- yy nn
        datapackage yy -- -- yy
        datafile    nn -- yy --

    """

    msg = "Conversion from {} to {} is not yet implemented".format(
        args.from_format, args.to_format
    )

    read_strategy = None
    write_strategy = None

    if args.from_format == "datafile":
        read_strategy = ReadDatafile()
    elif args.from_format == "datapackage":
        read_strategy = ReadDatapackage()
    elif args.from_format == "csv":
        read_strategy = ReadCsv()
    elif args.from_format == "excel":
        read_strategy = ReadExcel()

    if args.to_format == "datapackage":
        write_strategy = WriteDatapackage()
    elif args.to_format == "excel":
        write_strategy = WriteExcel()
    elif args.to_format == "datafile":
        write_strategy = WriteDatafile()
    elif args.to_format == "csv":
        write_strategy = WriteCsv()

    if read_strategy and write_strategy:
        context = Context(read_strategy, write_strategy)
        context.convert(args.from_path, args.to_path)
    else:
        raise NotImplementedError(msg)


def datapackage2res(args):
    create_res(args.datapackage, args.resfile)


def get_parser():
    parser = argparse.ArgumentParser(
        description="otoole: Python toolkit of OSeMOSYS users"
    )

    parser.add_argument(
        "--verbose", "-v", help="Enable debug mode", action="count", default=0
    )
    parser.add_argument(
        "--version",
        "-V",
        action="version",
        version=__version__,
        help="The version of otoole",
    )

    subparsers = parser.add_subparsers()

    # Parser for results
    result_parser = subparsers.add_parser("results", help="Post-process solution files")
    result_parser.add_argument(
        "from_format",
        help="Result data format to convert from",
        choices=sorted(["cbc", "cplex"]),
    )
    result_parser.add_argument(
        "to_format", help="Result data format to convert to", choices=sorted(["csv"]),
    )
    result_parser.add_argument(
        "from_path", help="Path to file or folder to convert from"
    )
    result_parser.add_argument("to_path", help="Path to file or folder to convert to")
    result_parser.add_argument(
        "--input_datapackage",
        help="Input data package required for OSeMOSYS short or fast results",
        default=None,
    )
    result_parser.add_argument(
        "--input_datafile",
        help="Input GNUMathProg datafile required for OSeMOSYS short or fast results",
        default=None,
    )
    result_parser.set_defaults(func=result_matrix)

    # Parser for conversion
    convert_parser = subparsers.add_parser(
        "convert", help="Convert from one input format to another"
    )
    convert_parser.add_argument(
        "from_format",
        help="Input data format to convert from",
        choices=sorted(["datafile", "datapackage", "excel", "csv"]),
    )
    convert_parser.add_argument(
        "to_format",
        help="Input data format to convert to",
        choices=sorted(["datafile", "datapackage", "csv", "excel"]),
    )
    convert_parser.add_argument(
        "from_path", help="Path to file or folder to convert from"
    )
    convert_parser.add_argument("to_path", help="Path to file or folder to convert to")
    convert_parser.set_defaults(func=conversion_matrix)

    # Parser for the CPLEX related commands
    cplex_parser = subparsers.add_parser("cplex", help="Process a CPLEX solution file")

    cplex_parser.add_argument(
        "cplex_file", help="The filepath of the OSeMOSYS cplex output file"
    )
    cplex_parser.add_argument(
        "output_file", help="The filepath of the converted file that will be written"
    )
    cplex_parser.add_argument(
        "-s",
        "--start_year",
        type=int,
        default=2015,
        help="Output only the results from this year onwards",
    )
    cplex_parser.add_argument(
        "-e",
        "--end_year",
        type=int,
        default=2070,
        help="Output only the results upto and including this year",
    )
    cplex_parser.add_argument("output_format", choices=["csv", "cbc"], default="cbc")
    cplex_parser.set_defaults(func=cplex2cbc)

    # Parser for validation
    valid_parser = subparsers.add_parser("validate", help="Validate an OSeMOSYS model")
    valid_parser.add_argument(
        "format",
        help="The format of the OSeMOSYS model to validate",
        choices=["datapackage", "sql"],
    )
    valid_parser.add_argument(
        "filepath", help="Path to the OSeMOSYS datapackage.json file to validate"
    )
    valid_parser.add_argument(
        "--config", help="Path to a user-defined validation-config file"
    )
    valid_parser.set_defaults(func=validate_model)

    # Parser for visualisation
    viz_parser = subparsers.add_parser("viz", help="Visualise the model")
    viz_subparsers = viz_parser.add_subparsers()

    res_parser = viz_subparsers.add_parser(
        "res", help="Generate a reference energy system"
    )
    res_parser.add_argument("datapackage", help="Path to model datapackage")
    res_parser.add_argument("resfile", help="Path to reference energy system")
    res_parser.set_defaults(func=datapackage2res)

    return parser


def main():

    parser = get_parser()
    args = parser.parse_args(sys.argv[1:])

    if args.verbose >= 1 and args.verbose < 2:
        logging.basicConfig(level=logging.INFO)
    if args.verbose > 2:
        logging.basicConfig(level=logging.DEBUG)

    def exception_handler(
        exception_type, exception, traceback, debug_hook=sys.excepthook
    ):
        if args.verbose:
            debug_hook(exception_type, exception, traceback)
        else:
            print("{}: {}".format(exception_type.__name__, exception.message))

    sys.excepthook = exception_handler

    if "func" in args:
        args.func(args)
    else:
        parser.print_help()



import pytest

from yaml import FullLoader, load

from otoole.validate import (
    compose_expression,
    compose_multi_expression,
    create_schema,
    read_validation_config,
    validate,
)


@pytest.mark.parametrize(
    "name,expected",
    [
        ("DZAETH", True),
        ("AGOCR1", True),
        ("CO1AGO", False),
        ("AGOETHETH", False),
        ("   ETH", False),
        ("DVA", False),
    ],
)
def test_validate_fuel_code_true(name, expected):

    actual = validate("^(DZA|AGO)(ETH|CR1)", name)
    assert actual == expected


@pytest.mark.parametrize(
    "name,expected",
    [
        ("DZAETH", True),
        ("AGOCR1", True),
        ("CO1AGO", True),
        ("AGOETHETH", False),
        ("   ETH", False),
        ("DVA", False),
    ],
)
def test_validate_fuel_code_true_multi(name, expected):

    actual = validate("^(DZA|AGO)(ETH|CR1)|^(CO1)(AGO)", name)
    assert actual == expected


def test_compose_expression():

    schema = [
        {"name": "countries", "valid": ["DZA", "AGO"], "position": (1, 3)},
        {"name": "fuels", "valid": ["ETH", "CR1"], "postion": (4, 6)},
    ]

    actual = compose_expression(schema)
    expected = "^(DZA|AGO)(ETH|CR1)"
    assert actual == expected


def test_compose_multi_expression():
    resource = load(
        """
- name: technology_name
  items:
  - name: countries
    valid:
      - DZA
      - AGO
    position: (1, 3)
  - name: fuels
    valid:
      - ETH
      - CR1
    position: (4, 6)
- name: trade_link
  items:
  - name: countries
    valid:
      - DZA
      - AGO
    position: (1, 3)
  - name: tradelink
    valid:
      - EL1EX
      - NG1EX
    position: (4, 8)
      """,
        Loader=FullLoader,
    )
    actual = compose_multi_expression(resource)
    expected = "^(DZA|AGO)(ETH|CR1)|^(DZA|AGO)(EL1EX|NG1EX)"
    assert actual == expected


def test_read_packaged_validation():

    actual = read_validation_config()
    expected = ["codes", "schema"]
    assert list(actual.keys()) == expected
    expected_codes = [
        "fuels",
        "technologies",
        "cooling",
        "tradelink",
        "age",
        "countries",
    ]
    assert list(actual["codes"].keys()) == expected_codes
    assert list(actual["codes"]["technologies"].keys()) == [
        "CH",
        "SC",
        "CV",
        "GC",
        "LS",
        "MS",
        "SS",
        "SA",
        "RC",
        "CC",
        "PW",
        "CN",
        "CS",
        "PU",
        "PV",
        "PS",
        "ON",
        "OF",
        "IM",
        "PR",
        "EX",
        "TR",
        "DI",
    ]


def test_create_schema():

    schema = {
        "codes": {"countries": {"DZA": "Algeria", "AGO": "Angola"}},
        "schema": {
            "FUEL": [
                {
                    "items": [
                        {"name": "countries", "valid": "countries", "position": (1, 3)}
                    ],
                    "name": "fuels",
                }
            ]
        },
    }
    actual = create_schema(schema)
    expected = {
        "FUEL": [
            {
                "items": [
                    {"name": "countries", "valid": ["DZA", "AGO"], "position": (1, 3)}
                ],
                "name": "fuels",
            }
        ]
    }
    assert actual == expected


def test_create_schema_two_items():

    schema = {
        "codes": {
            "countries": {"DZA": "Algeria", "AGO": "Angola"},
            "other_countries": {"ELC": "Electricity"},
        },
        "schema": {
            "FUEL": [
                {
                    "items": [
                        {"name": "countries", "valid": "countries", "position": (1, 3)}
                    ],
                    "name": "countries",
                },
                {
                    "items": [
                        {
                            "name": "other_countries",
                            "valid": "other_countries",
                            "position": (1, 3),
                        }
                    ],
                    "name": "other_countries",
                },
            ]
        },
    }
    actual = create_schema(schema)
    expected = {
        "FUEL": [
            {
                "items": [
                    {"name": "countries", "valid": ["DZA", "AGO"], "position": (1, 3)}
                ],
                "name": "countries",
            },
            {
                "items": [
                    {"name": "other_countries", "valid": ["ELC"], "position": (1, 3)}
                ],
                "name": "other_countries",
            },
        ]
    }
    assert actual == expected


def test_create_schema_duplicate_raises():

    schema = {
        "schema": {
            "FUEL": [
                {
                    "items": [
                        {
                            "name": "countries",
                            "valid": ["DZA", "DZA"],
                            "position": (1, 3),
                        }
                    ],
                    "name": "country",
                }
            ]
        }
    }
    with pytest.raises(ValueError):
        create_schema(schema)



import tempfile

from otoole.visualise import create_res

url = "tests/fixtures/simplicity-v0.2.1.zip"


def test_create_res():

    _, path_to_resfile = tempfile.mkstemp(suffix=".pdf")
    create_res(url, path_to_resfile)



import logging
from typing import Any, Dict, List, Optional, Set, TextIO, Tuple, Union

import pandas as pd

from otoole.input import ReadStrategy
from otoole.preprocess.longify_data import check_datatypes
from otoole.results.result_package import ResultsPackage

LOGGER = logging.getLogger(__name__)


class ReadCplex(ReadStrategy):
    """
    """

    def read(
        self, filepath: Union[str, TextIO], **kwargs
    ) -> Tuple[Dict[str, pd.DataFrame], Dict[str, Any]]:
        data = {}  # type: Dict

        if "input_data" in kwargs:
            input_data = kwargs["input_data"]
            years = input_data["YEAR"].values  # type: List
            start_year = int(years[0])
            end_year = int(years[-1])
        else:
            raise RuntimeError("To process CPLEX results please provide the input file")

        for linenum, line in enumerate(filepath):
            try:
                row_as_list = line.split("\t")
                name, df = self.convert_df(row_as_list, start_year, end_year,)

                if name in data:
                    data[name] = data[name].append(df)
                else:
                    data[name] = [df]

            except ValueError as ex:
                msg = "Error caused at line {}: {}. {}"
                raise ValueError(msg.format(linenum, line, ex))

        results = {}

        for name, contents in data.items():
            results[name] = pd.concat(contents)

        return results, self._read_default_values(self.results_config)

    def extract_variable_dimensions_values(self, data: List) -> Tuple[str, Tuple, List]:
        """Extracts useful information from a line of a results file
        """
        variable = data[0]
        number = len(self.results_config[variable]["indices"])
        dimensions = tuple(data[1:(number)])
        values = data[(number):]
        return (variable, dimensions, values)

    def convert_df(
        self, row_as_list: List, start_year: int, end_year: int
    ) -> Tuple[str, pd.DataFrame]:
        """Read the cplex line into a pandas DataFrame

        """
        variable, dimensions, values = self.extract_variable_dimensions_values(
            row_as_list
        )
        index = self.results_config[variable]["indices"]
        columns = index[:-1] + list(range(start_year, end_year + 1, 1))
        df = pd.DataFrame(data=[list(dimensions) + values], columns=columns).set_index(
            index[:-1]
        )
        df = df.melt(var_name="YEAR", value_name="VALUE", ignore_index=False)
        df = df.reset_index()
        df = check_datatypes(df, {**self.input_config, **self.results_config}, variable)
        df = df.set_index(index)
        df = df[(df != 0).any(axis=1)]
        return (variable, df)


class ReadCbc(ReadStrategy):
    """Read a CBC solution file into memory

    Arguments
    ---------
    user_config
    results_config
    """

    def read(
        self, filepath: Union[str, TextIO], **kwargs
    ) -> Tuple[Dict[str, pd.DataFrame], Dict[str, Any]]:
        """Read a solution file from ``filepath`` and process using ``input_data``

        Arguments
        ---------
        filepath : str, TextIO
            A path name or file buffer pointing to the CBC solution file
        input_data : dict, default=None
            dict of dataframes

        Returns
        -------
        tuple
            A tuple containing dict of pandas.DataFrames and a dict of default_values

        """
        if "input_data" in kwargs:
            input_data = kwargs["input_data"]
        else:
            input_data = None

        cbc = self._convert_cbc_to_dataframe(filepath)
        results = self._convert_dataframe_to_csv(cbc, input_data)
        default_values = self._read_default_values(self.results_config)
        return results, default_values

    def _convert_cbc_to_dataframe(self, file_path: Union[str, TextIO]) -> pd.DataFrame:
        """Reads a CBC solution file into a pandas DataFrame

        Arguments
        ---------
        file_path : str
        """
        df = pd.read_csv(
            file_path,
            header=None,
            names=["temp", "VALUE"],
            delim_whitespace=True,
            skiprows=1,
            usecols=[1, 2],
        )  # type: pd.DataFrame
        df.columns = ["temp", "Value"]
        df[["Variable", "Index"]] = df["temp"].str.split("(", expand=True)
        df = df.drop("temp", axis=1)
        df["Index"] = df["Index"].str.replace(")", "")
        return df[["Variable", "Index", "Value"]]

    def _convert_dataframe_to_csv(
        self, data: pd.DataFrame, input_data: Optional[Dict[str, pd.DataFrame]] = None
    ) -> Dict[str, pd.DataFrame]:
        """Convert from dataframe to long format

        Converts a pandas DataFrame containing all CBC results to reformatted
        dictionary of pandas DataFrames in long format ready to write out as
        csv files

        Arguments
        ---------
        data : pandas.DataFrame
            CBC results stored in a dataframe
        input_data_path : str, default=None
            Path to the OSeMOSYS data file containing input data

        Example
        -------
        >>> df = pd.DataFrame(data=[
                ['TotalDiscountedCost', "SIMPLICITY,2015", 187.01576],
                ['TotalDiscountedCost', "SIMPLICITY,2016", 183.30788]],
                columns=['Variable', 'Index', 'Value'])
        >>> convert_dataframe_to_csv(df)
        {'TotalDiscountedCost':        REGION  YEAR      VALUE
                                    0  SIMPLICITY  2015  187.01576
                                    1  SIMPLICITY  2016  183.30788}
        """
        input_config = self.input_config
        results_config = self.results_config

        sets = {x: y for x, y in input_config.items() if y["type"] == "set"}

        results = {}  # type: Dict[str, pd.DataFrame]

        not_found = []

        for name, details in sorted(results_config.items()):
            df = data[data["Variable"] == name]

            if not df.empty:

                LOGGER.debug("Extracting results for %s", name)
                indices = details["indices"]

                df[indices] = df["Index"].str.split(",", expand=True)

                types = {index: sets[index]["dtype"] for index in indices}
                df = df.astype(types)

                df = df.drop(columns=["Variable", "Index"])

                df = df.rename(columns={"Value": "VALUE"})

                columns = indices + ["VALUE"]

                df = df[columns]

                index = details["indices"].copy()
                # catches pandas error when there are duplicate column indices
                if check_duplicate_index(index):
                    index = rename_duplicate_column(index)
                    LOGGER.debug("Original column names: %s", columns)
                    renamed_columns = rename_duplicate_column(columns)
                    LOGGER.debug("New column names: %s", renamed_columns)
                    df.columns = renamed_columns
                results[name] = df.set_index(index)
            else:
                not_found.append(name)

        LOGGER.debug("Unable to find CBC variables for: %s", ", ".join(not_found))

        results_package = ResultsPackage(results, input_data)

        for name in not_found:

            LOGGER.info("Looking for %s", name)
            details = results_config[name]

            try:
                df = results_package[name]
            except KeyError as ex:
                LOGGER.info("No calculation method available for %s", name)
                LOGGER.debug("Error calculating %s: %s", name, str(ex))
                df = pd.DataFrame()

            if not df.empty:
                results[name] = df
            else:
                LOGGER.warning(
                    "Calculation returned empty dataframe for parameter '%s'", name
                )

        return results


def check_duplicate_index(index: List) -> bool:
    return len(set(index)) != len(index)


def identify_duplicate(index: List) -> Union[int, bool]:
    elements = set()  # type: Set
    for counter, elem in enumerate(index):
        if elem in elements:
            return counter
        else:
            elements.add(elem)
    return False


def rename_duplicate_column(index: List) -> List:
    column = index.copy()
    location = identify_duplicate(column)
    if location:
        column[location] = "_" + column[location]
    return column



from io import StringIO
from textwrap import dedent
from typing import List

from pytest import mark

import pandas as pd
from amply import Amply

from otoole import ReadDatafile, ReadMemory
from otoole.results.results import (
    ReadCbc,
    ReadCplex,
    check_duplicate_index,
    identify_duplicate,
    rename_duplicate_column,
)


class TestReadCplex:

    cplex_empty = "AnnualFixedOperatingCost	REGION	AOBACKSTOP	0	0	0	0	0	0	0	0	0	0	0	0	0	0	0	0	0	0	0	0	0	0	0	0	0	0	0	0	0	0	0	0	0	0	0	0	0	0	0	0	0	0	0	0	0	0	0	0	0	0	0	0	0	0	0	0"
    cplex_short = "AnnualFixedOperatingCost	REGION	CDBACKSTOP	0.0	0.0	137958.8400384134	305945.38410619126	626159.9611543404	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0"
    cplex_long = "RateOfActivity	REGION	S1D1	CGLFRCFURX	1	0.0	0.0	0.0	0.0	0.0	0.3284446367303371	0.3451714779880536	0.3366163200621617	0.3394945166233896	0.3137488154250392	0.28605725055560716	0.2572505015401749	0.06757558148965725	0.0558936625751148	0.04330608461292407	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0"

    cplex_mid_short = (
        "AnnualFixedOperatingCost",
        pd.DataFrame(
            data=[
                ["REGION", "CDBACKSTOP", 2017, 137958.8400384134],
                ["REGION", "CDBACKSTOP", 2018, 305945.38410619126],
                ["REGION", "CDBACKSTOP", 2019, 626159.9611543404],
            ],
            columns=["REGION", "TECHNOLOGY", "YEAR", "VALUE"],
        ).set_index(["REGION", "TECHNOLOGY", "YEAR"]),
    )

    cplex_mid_long = (
        "RateOfActivity",
        pd.DataFrame(
            data=[
                ["REGION", "S1D1", "CGLFRCFURX", 1, 2020, 0.3284446367303371],
                ["REGION", "S1D1", "CGLFRCFURX", 1, 2021, 0.3451714779880536],
                ["REGION", "S1D1", "CGLFRCFURX", 1, 2022, 0.3366163200621617],
                ["REGION", "S1D1", "CGLFRCFURX", 1, 2023, 0.3394945166233896],
                ["REGION", "S1D1", "CGLFRCFURX", 1, 2024, 0.3137488154250392],
                ["REGION", "S1D1", "CGLFRCFURX", 1, 2025, 0.28605725055560716],
                ["REGION", "S1D1", "CGLFRCFURX", 1, 2026, 0.2572505015401749],
                ["REGION", "S1D1", "CGLFRCFURX", 1, 2027, 0.06757558148965725],
                ["REGION", "S1D1", "CGLFRCFURX", 1, 2028, 0.0558936625751148],
                ["REGION", "S1D1", "CGLFRCFURX", 1, 2029, 0.04330608461292407],
            ],
            columns=[
                "REGION",
                "TIMESLICE",
                "TECHNOLOGY",
                "MODE_OF_OPERATION",
                "YEAR",
                "VALUE",
            ],
        ).set_index(["REGION", "TIMESLICE", "TECHNOLOGY", "MODE_OF_OPERATION", "YEAR"]),
    )

    dataframe_short = {
        "AnnualFixedOperatingCost": pd.DataFrame(
            data=[
                ["REGION", "CDBACKSTOP", 2017, 137958.8400384134],
                ["REGION", "CDBACKSTOP", 2018, 305945.3841061913],
                ["REGION", "CDBACKSTOP", 2019, 626159.9611543404],
            ],
            columns=["REGION", "TECHNOLOGY", "YEAR", "VALUE"],
        ).set_index(["REGION", "TECHNOLOGY", "YEAR"])
    }

    dataframe_long = {
        "RateOfActivity": pd.DataFrame(
            data=[
                ["REGION", "S1D1", "CGLFRCFURX", 1, 2020, 0.3284446367303371],
                ["REGION", "S1D1", "CGLFRCFURX", 1, 2021, 0.3451714779880536],
                ["REGION", "S1D1", "CGLFRCFURX", 1, 2022, 0.3366163200621617],
                ["REGION", "S1D1", "CGLFRCFURX", 1, 2023, 0.3394945166233896],
                ["REGION", "S1D1", "CGLFRCFURX", 1, 2024, 0.3137488154250392],
                ["REGION", "S1D1", "CGLFRCFURX", 1, 2025, 0.28605725055560716],
                ["REGION", "S1D1", "CGLFRCFURX", 1, 2026, 0.2572505015401749],
                ["REGION", "S1D1", "CGLFRCFURX", 1, 2027, 0.06757558148965725],
                ["REGION", "S1D1", "CGLFRCFURX", 1, 2028, 0.0558936625751148],
                ["REGION", "S1D1", "CGLFRCFURX", 1, 2029, 0.04330608461292407],
            ],
            columns=[
                "REGION",
                "TIMESLICE",
                "TECHNOLOGY",
                "MODE_OF_OPERATION",
                "YEAR",
                "VALUE",
            ],
        ).set_index(["REGION", "TIMESLICE", "TECHNOLOGY", "MODE_OF_OPERATION", "YEAR"])
    }

    test_data = [
        (cplex_short, dataframe_short),
        (cplex_long, dataframe_long),
    ]

    @mark.parametrize("cplex_input,expected", test_data, ids=["short", "long"])
    def test_read_cplex_to_dataframe(self, cplex_input, expected):
        cplex_reader = ReadCplex()

        input_data = {
            "YEAR": pd.DataFrame(data=list(range(2015, 2071, 1)), columns=["VALUE"])
        }

        with StringIO(cplex_input) as file_buffer:
            actual, _ = cplex_reader.read(file_buffer, input_data=input_data)
        for name, item in actual.items():
            pd.testing.assert_frame_equal(item, expected[name])

    test_data_mid = [(cplex_short, cplex_mid_short), (cplex_long, cplex_mid_long)]

    def test_read_empty_cplex_to_dataframe(self):
        cplex_input = self.cplex_empty

        cplex_reader = ReadCplex()

        input_data = {
            "YEAR": pd.DataFrame(data=list(range(2015, 2071, 1)), columns=["VALUE"])
        }

        with StringIO(cplex_input) as file_buffer:
            data, _ = cplex_reader.read(file_buffer, input_data=input_data)
        assert "AnnualFixedOperatingCost" in data
        expected = (
            pd.DataFrame(columns=["REGION", "TECHNOLOGY", "YEAR", "VALUE"],)
            .astype({"VALUE": float, "YEAR": int})
            .set_index(["REGION", "TECHNOLOGY", "YEAR"])
        )
        actual = data["AnnualFixedOperatingCost"]
        pd.testing.assert_frame_equal(actual, expected)

    test_data_mid = [(cplex_short, cplex_mid_short), (cplex_long, cplex_mid_long)]

    @mark.parametrize("cplex_input,expected", test_data_mid, ids=["short", "long"])
    def test_convert_cplex_to_df(self, cplex_input, expected):

        data = cplex_input.split("\t")
        cplex_reader = ReadCplex()
        actual = cplex_reader.convert_df(data, 2015, 2070)
        pd.testing.assert_frame_equal(actual[1], expected[1])


class TestReadCbc:

    cbc_data = dedent(
        """0 Trade(Globe,Globe,IP,L_AGR,2015) -0.0 0
0 Trade(Globe,Globe,IP,L_AGR,2016) -1.0 0
0 Trade(Globe,Globe,IP,L_AGR,2017) -2.0 0
0 Trade(Globe,Globe,IP,L_AGR,2018) -3.0 0
0 Trade(Globe,Globe,IP,L_AGR,2019) -4.0 0
0 Trade(Globe,Globe,IP,L_AGR,2020) -5.0 0
            """
    )
    otoole_data = pd.DataFrame(
        data=[
            ["Globe", "Globe", "IP", "L_AGR", 2016, -1.0],
            ["Globe", "Globe", "IP", "L_AGR", 2017, -2.0],
            ["Globe", "Globe", "IP", "L_AGR", 2018, -3.0],
            ["Globe", "Globe", "IP", "L_AGR", 2019, -4.0],
            ["Globe", "Globe", "IP", "L_AGR", 2020, -5.0],
        ],
        columns=["REGION", "_REGION", "TIMESLICE", "FUEL", "YEAR", "VALUE"],
    ).set_index(["REGION", "_REGION", "TIMESLICE", "FUEL", "YEAR"])

    test_data = [(cbc_data, otoole_data,)]

    @mark.parametrize("cbc_input,expected", test_data)
    def test_read_cbc_to_otoole_dataframe(self, cbc_input, expected):
        with StringIO(cbc_input) as file_buffer:
            actual = ReadCbc().read(file_buffer, kwargs={"input_data": {}})[0]["Trade"]
        pd.testing.assert_frame_equal(actual, expected)

    def test_read_cbc_dataframe_to_otoole_dataframe(self):

        prelim_data = pd.DataFrame(
            data=[
                ["Trade", "Globe,Globe,IP,L_AGR,2016", -1.0],
                ["Trade", "Globe,Globe,IP,L_AGR,2017", -2.0],
                ["Trade", "Globe,Globe,IP,L_AGR,2018", -3.0],
                ["Trade", "Globe,Globe,IP,L_AGR,2019", -4.0],
                ["Trade", "Globe,Globe,IP,L_AGR,2020", -5.0],
            ],
            columns=["Variable", "Index", "Value"],
        )
        actual = ReadCbc()._convert_dataframe_to_csv(prelim_data, {})["Trade"]
        pd.testing.assert_frame_equal(actual, self.otoole_data)

    test_data_4 = [
        (["REGION", "REGION", "TIMESLICE", "FUEL", "YEAR"], True),
        (["REGION", "TIMESLICE", "FUEL", "YEAR"], False),
        ([], False),
    ]

    @mark.parametrize("data,expected", test_data_4)
    def test_handle_duplicate_indices(self, data, expected):
        assert check_duplicate_index(data) is expected

    test_data_5 = [
        (["REGION", "REGION", "TIMESLICE", "FUEL", "YEAR"], 1),
        (["REGION", "TIMESLICE", "FUEL", "YEAR"], False),
        ([], False),
    ]

    @mark.parametrize("data,expected", test_data_5)
    def test_identify_duplicate(self, data, expected):
        assert identify_duplicate(data) == expected

    def test_rename_duplicate_column(self):
        data = ["REGION", "REGION", "TIMESLICE", "FUEL", "YEAR"]
        actual = rename_duplicate_column(data)
        expected = ["REGION", "_REGION", "TIMESLICE", "FUEL", "YEAR"]
        assert actual == expected

    total_cost_cbc = dedent(
        """Optimal - objective value 4483.96932237
                             1 TotalDiscountedCost(SIMPLICITY,2015){0}187.01576{1}0
                             2 TotalDiscountedCost(SIMPLICITY,2016){0}183.30788{1}0
                             3 TotalDiscountedCost(SIMPLICITY,2017){0}181.05465{1}0
                             4 TotalDiscountedCost(SIMPLICITY,2018){0}218.08923{1}0
                             5 TotalDiscountedCost(SIMPLICITY,2019){0}193.85792{1}0
                             6 TotalDiscountedCost(SIMPLICITY,2020){0}233.79202{1}0

                         """.format(
            " " * 51, " " * 23
        )
    )

    total_cost_cbc_mid = pd.DataFrame(
        data=[
            ["TotalDiscountedCost", "SIMPLICITY,2015", 187.01576],
            ["TotalDiscountedCost", "SIMPLICITY,2016", 183.30788],
            ["TotalDiscountedCost", "SIMPLICITY,2017", 181.05465],
            ["TotalDiscountedCost", "SIMPLICITY,2018", 218.08923],
            ["TotalDiscountedCost", "SIMPLICITY,2019", 193.85792],
            ["TotalDiscountedCost", "SIMPLICITY,2020", 233.79202],
        ],
        columns=["Variable", "Index", "Value"],
    )

    total_cost_otoole_df = {
        "TotalDiscountedCost": pd.DataFrame(
            data=[
                ["SIMPLICITY", 2015, 187.01576],
                ["SIMPLICITY", 2016, 183.30788],
                ["SIMPLICITY", 2017, 181.05465],
                ["SIMPLICITY", 2018, 218.08923],
                ["SIMPLICITY", 2019, 193.85792],
                ["SIMPLICITY", 2020, 233.79202],
            ],
            columns=["REGION", "YEAR", "VALUE"],
        ).set_index(["REGION", "YEAR"])
    }

    test_data = [(total_cost_cbc, total_cost_cbc_mid)]

    @mark.parametrize("cbc_input,expected", test_data, ids=["TotalDiscountedCost"])
    def test_read_cbc_to_dataframe(self, cbc_input, expected):
        cbc_reader = ReadCbc()
        with StringIO(cbc_input) as file_buffer:
            actual = cbc_reader._convert_cbc_to_dataframe(file_buffer)
        pd.testing.assert_frame_equal(actual, expected)

    test_data_2 = [
        # First case
        (total_cost_cbc_mid, {}, total_cost_otoole_df),
        # Second case
        (
            pd.DataFrame(
                data=[
                    ["AnnualEmissions", "REGION,CO2,2017", 137958.8400384134],
                    ["AnnualEmissions", "REGION,CO2,2018", 305945.3841061913],
                    ["AnnualEmissions", "REGION,CO2,2019", 626159.9611543404],
                ],
                columns=["Variable", "Index", "Value"],
            ),
            {},
            {
                "AnnualEmissions": pd.DataFrame(
                    data=[
                        ["REGION", "CO2", 2017, 137958.8400384134],
                        ["REGION", "CO2", 2018, 305945.3841061913],
                        ["REGION", "CO2", 2019, 626159.9611543404],
                    ],
                    columns=["REGION", "EMISSION", "YEAR", "VALUE"],
                ).set_index(["REGION", "EMISSION", "YEAR"])
            },
        ),
    ]  # type: List

    @mark.parametrize(
        "results,cbc_input,expected",
        test_data_2,
        ids=["TotalDiscountedCost", "AnnualEmissions1"],
    )
    def test_convert_cbc_to_csv_long(self, results, cbc_input, expected):
        cbc_reader = ReadCbc()
        actual = cbc_reader._convert_dataframe_to_csv(results, cbc_input)
        assert isinstance(actual, dict)
        for name, df in actual.items():
            pd.testing.assert_frame_equal(df, expected[name])

    test_data_3 = [(total_cost_cbc, {}, total_cost_otoole_df)]  # type: List

    @mark.parametrize(
        "cbc_solution,input_data,expected", test_data_3, ids=["TotalDiscountedCost"],
    )
    def test_convert_cbc_to_csv_long_read(self, cbc_solution, input_data, expected):
        cbc_reader = ReadCbc()
        with StringIO(cbc_solution) as file_buffer:
            actual = cbc_reader.read(file_buffer, kwargs={"input_data": input_data})[0][
                "TotalDiscountedCost"
            ]
        assert isinstance(actual, pd.DataFrame)
        pd.testing.assert_frame_equal(actual, expected["TotalDiscountedCost"])

    def test_convert_cbc_to_csv_short(self):
        cbc_results = pd.DataFrame(
            data=[
                ["RateOfActivity", "SIMPLICITY,ID,GAS_EXTRACTION,1,2014", 1],
                ["RateOfActivity", "SIMPLICITY,IN,GAS_EXTRACTION,1,2014", 1],
                ["RateOfActivity", "SIMPLICITY,SD,GAS_EXTRACTION,1,2014", 1],
                ["RateOfActivity", "SIMPLICITY,SN,GAS_EXTRACTION,1,2014", 1],
                ["RateOfActivity", "SIMPLICITY,WD,GAS_EXTRACTION,1,2014", 1],
                ["RateOfActivity", "SIMPLICITY,WN,GAS_EXTRACTION,1,2014", 1],
                ["RateOfActivity", "SIMPLICITY,ID,DUMMY,1,2014", 1],
                ["RateOfActivity", "SIMPLICITY,IN,DUMMY,1,2014", 1],
                ["RateOfActivity", "SIMPLICITY,SD,DUMMY,1,2014", 1],
                ["RateOfActivity", "SIMPLICITY,SN,DUMMY,1,2014", 1],
                ["RateOfActivity", "SIMPLICITY,WD,DUMMY,1,2014", 1],
                ["RateOfActivity", "SIMPLICITY,WN,DUMMY,1,2014", 1],
            ],
            columns=["Variable", "Index", "Value"],
        )
        input_data = {
            "EmissionActivityRatio": pd.DataFrame(
                data=[["SIMPLICITY", "GAS_EXTRACTION", "CO2", 1, 2014, 1.0]],
                columns=[
                    "REGION",
                    "TECHNOLOGY",
                    "EMISSION",
                    "MODE_OF_OPERATION",
                    "YEAR",
                    "VALUE",
                ],
            ).set_index(
                ["REGION", "TECHNOLOGY", "EMISSION", "MODE_OF_OPERATION", "YEAR"]
            ),
            "YearSplit": pd.DataFrame(
                data=[
                    ["ID", 2014, 0.1667],
                    ["IN", 2014, 0.0833],
                    ["SD", 2014, 0.1667],
                    ["SN", 2014, 0.0833],
                    ["WD", 2014, 0.3333],
                    ["WN", 2014, 0.1667],
                ],
                columns=["TIMESLICE", "YEAR", "VALUE"],
            ).set_index(["TIMESLICE", "YEAR"]),
        }

        expected = pd.DataFrame(
            data=[["SIMPLICITY", "CO2", 2014, 1.0]],
            columns=["REGION", "EMISSION", "YEAR", "VALUE"],
        ).set_index(["REGION", "EMISSION", "YEAR"])

        cbc_reader = ReadCbc()
        actual = cbc_reader._convert_dataframe_to_csv(cbc_results, input_data)
        assert isinstance(actual, dict)
        pd.testing.assert_frame_equal(actual["AnnualEmissions"], expected)

    def test_solution_to_dataframe(self):
        input_file = self.total_cost_cbc
        reader = ReadCbc()
        with StringIO(input_file) as file_buffer:
            actual = reader.read(file_buffer)
        expected = self.total_cost_otoole_df
        pd.testing.assert_frame_equal(
            actual[0]["TotalDiscountedCost"], expected["TotalDiscountedCost"]
        )


class TestReadMemoryStrategy:
    def test_read_memory(self):

        data = [
            ["SIMPLICITY", "ETH", 2014, 1.0],
            ["SIMPLICITY", "RAWSUG", 2014, 0.5],
            ["SIMPLICITY", "ETH", 2015, 1.03],
            ["SIMPLICITY", "RAWSUG", 2015, 0.51],
        ]
        df = pd.DataFrame(data=data, columns=["REGION", "FUEL", "YEAR", "VALUE"])
        parameters = {"AccumulatedAnnualDemand": df}

        actual, default_values = ReadMemory(parameters).read()

        expected = {
            "AccumulatedAnnualDemand": pd.DataFrame(
                data=data, columns=["REGION", "FUEL", "YEAR", "VALUE"]
            ).set_index(["REGION", "FUEL", "YEAR"])
        }

        assert "AccumulatedAnnualDemand" in actual.keys()
        pd.testing.assert_frame_equal(
            actual["AccumulatedAnnualDemand"], expected["AccumulatedAnnualDemand"]
        )


class TestConfig:
    def test_read_config(self):

        read = ReadDatafile()

        actual = read._read_config()
        expected = {
            "default": 0,
            "dtype": "float",
            "indices": ["REGION", "FUEL", "YEAR"],
            "type": "param",
        }
        assert actual["AccumulatedAnnualDemand"] == expected


class TestReadDatafile:
    def test_amply(self):

        Amply(
            """set REGION;
            # set REGION := SIMPLICITY;
            set TECHNOLOGY;
            set TECHNOLOGY := ETHPLANT GAS_EXTRACTION;
            set MODE_OF_OPERATION;
            set MODE_OF_OPERATION := 1 2;
            set YEAR;
            set YEAR := 2014;
            end;"""
        )

    def test_convert_amply_to_dataframe(self):

        config = {
            "VariableCost": {
                "type": "param",
                "indices": ["REGION", "TECHNOLOGY", "MODE_OF_OPERATION", "YEAR"],
                "dtype": "float",
                "default": 0,
            },
            "REGION": {"type": "set", "dtype": "str"},
            "YEAR": {"dtype": "int", "type": "set"},
            "MODE_OF_OPERATION": {"dtype": "int", "type": "set"},
            "TECHNOLOGY": {"dtype": "str", "type": "set"},
        }

        amply = Amply(
            """set REGION;
                        set REGION := SIMPLICITY;
                        set TECHNOLOGY;
                        set TECHNOLOGY := ETHPLANT GAS_EXTRACTION;
                        set MODE_OF_OPERATION;
                        set MODE_OF_OPERATION := 1 2;
                        set YEAR;
                        set YEAR := 2014;"""
        )
        amply.load_string(
            "param VariableCost {REGION,TECHNOLOGY,MODE_OF_OPERATION,YEAR};"
        )
        #     amply.load_string("""param default 0 : VariableCost :=
        # SIMPLICITY ETHPLANT 1 2014 2.89
        # SIMPLICITY ETHPLANT 2 2014 999999.0
        # SIMPLICITY GAS_EXTRACTION 1 2014 7.5
        # SIMPLICITY GAS_EXTRACTION 2 2014 999999.0""")
        amply.load_string(
            """
    param VariableCost default 0.0001 :=
    [SIMPLICITY,ETHPLANT,*,*]:
    2014 :=
    1 2.89
    2 999999.0
    [SIMPLICITY,GAS_EXTRACTION,*,*]:
    2014 :=
    1 7.5
    2 999999.0;"""
        )

        read = ReadDatafile()

        actual = read._convert_amply_to_dataframe(amply, config)
        expected = pd.DataFrame(
            data=[
                ["SIMPLICITY", "ETHPLANT", 1, 2014, 2.89],
                ["SIMPLICITY", "ETHPLANT", 2, 2014, 999999.0],
                ["SIMPLICITY", "GAS_EXTRACTION", 1, 2014, 7.5],
                ["SIMPLICITY", "GAS_EXTRACTION", 2, 2014, 999999.0],
            ],
            columns=["REGION", "TECHNOLOGY", "MODE_OF_OPERATION", "YEAR", "VALUE"],
        )
        pd.testing.assert_frame_equal(actual["VariableCost"], expected)

    def test_convert_amply_data_to_list_of_lists(self):

        data = {
            "SIMPLICITY": {
                "ETHPLANT": {1.0: {2014.0: 2.89}, 2.0: {2014.0: 999999.0}},
                "GAS_EXTRACTION": {1.0: {2014.0: 7.5}, 2.0: {2014.0: 999999.0}},
            }
        }
        expected = [
            ["SIMPLICITY", "ETHPLANT", 1.0, 2014.0, 2.89],
            ["SIMPLICITY", "ETHPLANT", 2.0, 2014.0, 999999.0],
            ["SIMPLICITY", "GAS_EXTRACTION", 1.0, 2014.0, 7.5],
            ["SIMPLICITY", "GAS_EXTRACTION", 2.0, 2014.0, 999999.0],
        ]
        read = ReadDatafile()
        actual = read._convert_amply_data_to_list(data)
        assert actual == expected

    def test_load_parameters(self):

        config = {"TestParameter": {"type": "param", "indices": ["index1", "index2"]}}
        read = ReadDatafile()
        actual = read._load_parameter_definitions(config)
        expected = "param TestParameter {index1,index2};\n"
        assert actual == expected

    def test_load_sets(self):

        config = {"TestSet": {"type": "set"}}

        read = ReadDatafile()
        actual = read._load_parameter_definitions(config)
        expected = "set TestSet;\n"
        assert actual == expected

    def test_catch_error_no_parameter(self, caplog):
        """Fix for https://github.com/OSeMOSYS/otoole/issues/70 where parameter in
        datafile but not in config causes error.  Instead, throw warning (and advise
        that user should use a custom configuration).
        """
        read = ReadDatafile()
        config = read.input_config
        amply_datafile = amply = Amply(
            """set REGION;
            set TECHNOLOGY;
            set MODE_OF_OPERATION;
            set YEAR;"""
        )
        amply.load_string("""param ResultsPath := 'test_path';""")
        read._convert_amply_to_dataframe(amply_datafile, config)
        assert (
            "Parameter ResultsPath could not be found in the configuration."
            in caplog.text
        )



import logging
from collections.abc import Mapping
from datetime import datetime
from typing import Dict, Iterable, List, Optional, Tuple

import pandas as pd

LOGGER = logging.getLogger(__name__)


class ResultsPackage(Mapping):
    """A package of OSeMOSYS results

    Internal data structure is a dictionary of pandas DataFrames

    A crude caching function is implemented whereby calculated results are stored in the
    internal data structure so that each result is only calculated once.

    Arguments
    ---------
    data : dict
        A dictionary of results data
    input_data: dict, default=None
        Dictionary of input data

    """

    def __init__(
        self,
        data: Dict[str, pd.DataFrame],
        input_data: Optional[Dict[str, pd.DataFrame]] = None,
    ):
        super().__init__()
        self._data = data
        if input_data:
            self._package = input_data
        else:
            self._package = {}
        self._result_mapper = {
            "AccumulatedNewCapacity": self.accumulated_new_capacity,
            "AnnualEmissions": self.annual_emissions,
            "AnnualFixedOperatingCost": self.annual_fixed_operating_cost,
            "AnnualTechnologyEmission": self.annual_technology_emissions,
            "AnnualTechnologyEmissionByMode": self.annual_technology_emission_by_mode,
            "AnnualVariableOperatingCost": self.annual_variable_operating_cost,
            "CapitalInvestment": self.capital_investment,
            "Demand": self.demand,
            "DiscountedTechnologyEmissionsPenalty": self.discounted_tech_emis_pen,
            "ProductionByTechnology": self.production_by_technology,
            "ProductionByTechnologyAnnual": self.production_by_technology_annual,
            "RateOfProductionByTechnology": self.rate_of_product_technology,
            "RateOfProductionByTechnologyByMode": self.rate_of_production_tech_mode,
            "RateOfUseByTechnology": self.rate_of_use_by_technology,
            "RateOfUseByTechnologyByMode": self.rate_of_use_by_technology_by_mode,
            "TotalAnnualTechnologyActivityByMode": self.total_annual_tech_activity_mode,
            "TotalCapacityAnnual": self.total_capacity_annual,
            "TotalDiscountedCost": self.total_discounted_cost,
            "TotalTechnologyAnnualActivity": self.total_technology_annual_activity,
            "TotalTechnologyModelPeriodActivity": self.total_tech_model_period_activity,
            "UseByTechnology": self.use_by_technology,
        }
        self._result_cache = {}  # type: Dict[str, pd.DataFrame]

    @property
    def data(self) -> Dict[str, pd.DataFrame]:
        """View the results dictionary
        """
        return self._data

    @property
    def result_mapper(self) -> Dict[str, pd.DataFrame]:
        return self._result_mapper

    @property
    def result_cache(self) -> Dict[str, pd.DataFrame]:
        return self._result_cache

    @result_cache.setter
    def result_cache(self, value: Iterable[Tuple[str, pd.DataFrame]]):
        self._result_cache.update(value)

    def __getitem__(self, name: str) -> pd.DataFrame:
        LOGGER.debug("Returning '%s' from ... ", name)
        if name in self.data.keys():
            LOGGER.debug("    ... ResultsPackage.data")
            return self.data[name]
        elif name in self._package.keys():
            LOGGER.debug("    ... ResultsPackage.input_data")
            return self._package[name]
        elif name in self.result_cache.keys():
            LOGGER.debug("    ... ResultsPackage.result_cache")
            return self.result_cache[name]
        elif name in self.result_mapper.keys():
            # Implements a crude form of caching, where calculated results are
            # first stored in the internal dict, and then returned
            LOGGER.debug("  ... ResultsPackage.calculating ...")
            start = datetime.now()
            results = self.result_mapper[name]()
            stop = datetime.now()
            diff = stop - start
            total = diff.total_seconds()
            LOGGER.debug("Calculation took %s secs", total)
            LOGGER.debug("Caching results for %s", name)
            self.result_cache[name] = results
            return self.result_cache[name]
        else:
            LOGGER.debug("  ... Not found in cache or calculation methods")
            raise KeyError("{} is not accessible or available".format(name))
        return self.data[name]

    def __iter__(self):
        raise NotImplementedError()

    def __len__(self):
        raise NotImplementedError()

    def _msg(self, name: str, error: str):
        return "Cannot calculate {} due to missing data: {}".format(name, error)

    def accumulated_new_capacity(self) -> pd.DataFrame:
        """AccumulatedNewCapacity

        Arguments
        ---------
        operational_life: pandas.DataFrame

        new_capacity: pandas.DataFrame

        year: pandas.Index

        Notes
        -----
        From the formulation::

            r~REGION, t~TECHNOLOGY, y~YEAR,
            sum{yy in YEAR: y-yy < OperationalLife[r,t] && y-yy>=0}
                NewCapacity[r,t,yy] ~VALUE;

        """
        try:
            new_capacity = self["NewCapacity"].copy()
            year = pd.Index(self["YEAR"]["VALUE"].to_list())
        except KeyError as ex:
            raise KeyError(self._msg("AccumulatedNewCapacity", str(ex)))

        new_capacity["OperationalLife"] = self["OperationalLife"].copy()

        regions = new_capacity.reset_index()["REGION"].unique()
        technologies = new_capacity.reset_index()["TECHNOLOGY"].unique()

        index = pd.MultiIndex.from_product(
            [regions, technologies, year.to_list()],
            names=["REGION", "TECHNOLOGY", "YEAR"],
        )

        acc_capacity = new_capacity.reindex(index, copy=True)

        for index, data in new_capacity.reset_index().groupby(
            by=["REGION", "TECHNOLOGY"]
        ):
            region, technology = index
            for yr in year:
                mask = (yr - data["YEAR"] < data["OperationalLife"]) & (
                    yr - data["YEAR"] >= 0
                )
                acc_capacity.loc[region, technology, yr] = data[mask].sum()

        acc_capacity = acc_capacity.drop(columns="OperationalLife")
        return acc_capacity[(acc_capacity != 0).all(1)]

    def annual_emissions(self) -> pd.DataFrame:
        """Calculates the annual emissions

        Annual Emission are found by multiplying the emission activity ratio
        (emissions per unit activity) by the rate of activity and the yearsplit

        Notes
        -----
        From the formulation::

            sum{t in TECHNOLOGY, l in TIMESLICE, m in MODE_OF_OPERATION:
                    EmissionActivityRatio[r,t,e,m,y]<>0}
                RateOfActivity[r,l,t,m,y] * EmissionActivityRatio[r,t,e,m,y]
                    * YearSplit[l,y]~VALUE;
        """
        try:
            emission_activity_ratio = self["EmissionActivityRatio"]
            yearsplit = self["YearSplit"]
            rate_of_activity = self["RateOfActivity"]
        except KeyError as ex:
            raise KeyError(self._msg("AnnualEmissions", str(ex)))

        mid = emission_activity_ratio.mul(yearsplit)
        data = mid.mul(rate_of_activity, fill_value=0.0)

        if not data.empty:
            data = data.groupby(by=["REGION", "EMISSION", "YEAR"]).sum()
        return data

    def annual_fixed_operating_cost(self) -> pd.DataFrame:
        """Compute AnnualFixedOperatingCost result

        Notes
        -----
        From the formulation::

            r~REGION, t~TECHNOLOGY, y~YEAR,
            FixedCost[r,t,y] *
            ((sum{yy in YEAR: y-yy < OperationalLife[r,t] && y-yy>=0}
                NewCapacity[r,t,yy]) + ResidualCapacity[r,t,y]) ~VALUE;

        """
        try:
            total_capacity = self["TotalCapacityAnnual"]
            fixed_cost = self["FixedCost"]
        except KeyError as ex:
            raise KeyError(self._msg("AnnualFixedOperatingCost", str(ex)))

        total_fixed_costs = total_capacity.mul(fixed_cost, fill_value=0.0)

        return total_fixed_costs[(total_fixed_costs != 0).all(1)].dropna()

    def annual_technology_emissions(self) -> pd.DataFrame:
        """Calculates results ``AnnualTechnologyEmission``

        Notes
        -----
        From the formulation::

            REGION, TECHNOLOGY, EMISSION, YEAR,
            sum{l in TIMESLICE, m in MODE_OF_OPERATION:
                EmissionActivityRatio[r,t,e,m,y]<>0}
            EmissionActivityRatio[r,t,e,m,y] * RateOfActivity[r,l,t,m,y]
                * YearSplit[l,y];

        """
        try:
            data = self["AnnualTechnologyEmissionByMode"].copy(deep=True)
        except KeyError as ex:
            raise KeyError(self._msg("AnnualTechnologyEmission", str(ex)))

        if not data.empty:
            data = data.groupby(by=["REGION", "TECHNOLOGY", "EMISSION", "YEAR"]).sum()

        return data[(data != 0).all(1)]

    def annual_technology_emission_by_mode(self) -> pd.DataFrame:
        """

        Notes
        -----
        From the formulation::

            r~REGION, t~TECHNOLOGY, e~EMISSION, m~MODE_OF_OPERATION, y~YEAR,
            sum{l in TIMESLICE: EmissionActivityRatio[r,t,e,m,y] <> 0}
                EmissionActivityRatio[r,t,e,m,y] * RateOfActivity[r,l,t,m,y]
                    * YearSplit[l,y]
        """
        try:
            emission_activity_ratio: pd.DataFrame = self["EmissionActivityRatio"]
            yearsplit: pd.DataFrame = self["YearSplit"]
            rate_of_activity: pd.DataFrame = self["RateOfActivity"]
        except KeyError as ex:
            raise KeyError(self._msg("AnnualTechnologyEmissionByMode", str(ex)))

        mid = emission_activity_ratio.mul(yearsplit)
        data = mid.mul(rate_of_activity)

        if not data.empty:
            data = data.groupby(
                by=["REGION", "TECHNOLOGY", "EMISSION", "MODE_OF_OPERATION", "YEAR"]
            ).sum()

        return data[(data != 0).all(1)]

    def annual_variable_operating_cost(self) -> pd.DataFrame:
        """AnnualVariableOperatingCost

        Notes
        -----
        From the formulation::

            r~REGION, t~TECHNOLOGY, y~YEAR,
            sum{m in MODE_OF_OPERATION, l in TIMESLICE}
                RateOfActivity[r,l,t,m,y]
                * YearSplit[l,y]
                * VariableCost[r,t,m,y] ~VALUE;

        """
        try:
            rate_of_activity = self["RateOfActivity"]
            yearsplit = self["YearSplit"]
            variable_cost = self["VariableCost"]
        except KeyError as ex:
            raise KeyError(self._msg("AnnualVariableOperatingCost", str(ex)))

        split_activity = rate_of_activity.mul(yearsplit, fill_value=0.0)
        data = split_activity.mul(variable_cost, fill_value=0.0)
        if not data.empty:
            data = data.groupby(by=["REGION", "TECHNOLOGY", "YEAR"]).sum()
        return data[(data != 0).all(1)]

    def capital_investment(self) -> pd.DataFrame:
        """CapitalInvestment

        Notes
        -----
        From the formulation::

            r~REGION, t~TECHNOLOGY, y~YEAR,
            CapitalCost[r,t,y] * NewCapacity[r,t,y] ~VALUE;
        """
        try:
            capital_cost = self["CapitalCost"]
            new_capacity = self["NewCapacity"]
        except KeyError as ex:
            raise KeyError(self._msg("CapitalInvestment", str(ex)))

        data = capital_cost.mul(new_capacity, fill_value=0.0)
        return data[(data != 0).all(1)]

    def demand(self) -> pd.DataFrame:
        """Demand

        Notes
        -----
        From the formulation::

            r~REGION, l~TIMESLICE, f~FUEL, y~YEAR,
            SpecifiedAnnualDemand[r,f,y] * SpecifiedDemandProfile[r,f,l,y] ~VALUE;
        """
        try:
            specified_annual_demand = self["SpecifiedAnnualDemand"]
            specified_demand_profile = self["SpecifiedDemandProfile"]
        except KeyError as ex:
            raise KeyError(self._msg("Demand", str(ex)))

        data = specified_annual_demand.mul(specified_demand_profile, fill_value=0.0)
        if not data.empty:
            data = data.reset_index().set_index(["REGION", "TIMESLICE", "FUEL", "YEAR"])
        return data[(data != 0).all(1)]

    def discounted_tech_emis_pen(self) -> pd.DataFrame:
        """
        Notes
        -----
        From the formulation::

            DiscountedTechnologyEmissionsPenalty[r,t,y] :=

            EmissionActivityRatio[r,t,e,m,y] * RateOfActivity[r,l,t,m,y] *
            YearSplit[l,y] * EmissionsPenalty[r,e,y] /
            ((1+DiscountRate[r, t]) ^ (y - min{yy in YEAR} min(yy) + 0.5))

        """
        try:
            annual_technology_emission_by_mode = self["AnnualTechnologyEmissionByMode"]
            emission_penalty = self["EmissionsPenalty"]
            regions = self["REGION"]["VALUE"].to_list()
            technologies = list(
                annual_technology_emission_by_mode.reset_index()["TECHNOLOGY"].unique()
            )
            years = self["YEAR"]["VALUE"].to_list()
            discount_rate = self["DiscountRate"]

            crf = capital_recovery_factor(
                regions, technologies, years, discount_rate, adj=0.5
            )
        except KeyError as ex:
            raise KeyError(self._msg("DiscountedTechnologyEmissionsPenalty", str(ex)))

        emissions_penalty = annual_technology_emission_by_mode.mul(
            emission_penalty, fill_value=0.0
        )
        data = emissions_penalty.div(crf, fill_value=0.0)

        if not data.empty:
            data = data.groupby(by=["REGION", "TECHNOLOGY", "YEAR"]).sum()

        return data[(data != 0).all(1)]

    def production_by_technology(self) -> pd.DataFrame:
        """Compute production by technology

        ProductionByTechnology

        Notes
        -----
        From the formulation::

            r~REGION, l~TIMESLICE, t~TECHNOLOGY, f~FUEL, y~YEAR,
            sum{m in MODE_OF_OPERATION: OutputActivityRatio[r,t,f,m,y] <> 0}
                RateOfActivity[r,l,t,m,y] * OutputActivityRatio[r,t,f,m,y]
                * YearSplit[l,y] ~VALUE;
        """
        try:
            rate_of_activity = self["RateOfActivity"]
            output_activity_ratio = self["OutputActivityRatio"]
            year_split = self["YearSplit"]
        except KeyError as ex:
            raise KeyError(self._msg("ProductionByTechnology", str(ex)))

        split_activity = rate_of_activity.mul(year_split, fill_value=0.0)
        data = split_activity.mul(output_activity_ratio, fill_value=0.0)
        if not data.empty:
            data = data.groupby(
                by=["REGION", "TIMESLICE", "TECHNOLOGY", "FUEL", "YEAR"]
            ).sum()
        return data[(data != 0).all(1)]

    def production_by_technology_annual(self) -> pd.DataFrame:
        """Aggregates production by technology to the annual level
        """
        try:
            production_by_technology = self["ProductionByTechnology"].copy(deep=True)
        except KeyError as ex:
            raise KeyError(self._msg("ProductionByTechnologyAnnual", str(ex)))

        data = production_by_technology
        if not data.empty:
            data = data.groupby(by=["REGION", "TECHNOLOGY", "FUEL", "YEAR"]).sum()
        return data[(data != 0).all(1)]

    def rate_of_production_tech_mode(self) -> pd.DataFrame:
        """RateOfProductionByTechnologyByMode

        Notes
        -----
        From the formulation::

            r~REGION, l~TIMESLICE, t~TECHNOLOGY, m~MODE_OF_OPERATION, f~FUEL, y~YEAR,
            RateOfActivity[r,l,t,m,y] * OutputActivityRatio[r,t,f,m,y]~VALUE;
        """
        try:
            rate_of_activity = self["RateOfActivity"]
            output_activity_ratio = self["OutputActivityRatio"]
        except KeyError as ex:
            raise KeyError(self._msg("RateOfProductionByTechnologyByMode", str(ex)))

        data = rate_of_activity.mul(output_activity_ratio, fill_value=0.0)
        if not data.empty:
            data = data.reset_index().set_index(
                [
                    "REGION",
                    "TIMESLICE",
                    "TECHNOLOGY",
                    "MODE_OF_OPERATION",
                    "FUEL",
                    "YEAR",
                ]
            )
        return data[(data != 0).all(1)].sort_index()

    def rate_of_product_technology(self) -> pd.DataFrame:
        """Sums up mode of operation for rate of production

        Notes
        -----
        From the formulation::

            r~REGION, l~TIMESLICE, t~TECHNOLOGY, f~FUEL, y~YEAR,
            sum{m in MODE_OF_OPERATION: OutputActivityRatio[r,t,f,m,y] <> 0}
                RateOfActivity[r,l,t,m,y] * OutputActivityRatio[r,t,f,m,y]~VALUE;

        """
        try:
            rate_of_production = self["RateOfProductionByTechnologyByMode"].copy(
                deep=True
            )
        except KeyError as ex:
            raise KeyError(self._msg("RateOfProductionByTechnology", str(ex)))

        data = rate_of_production
        if not data.empty:
            data = data.groupby(
                by=["REGION", "TIMESLICE", "TECHNOLOGY", "FUEL", "YEAR"]
            ).sum()
        return data[(data != 0).all(1)].sort_index()

    def rate_of_use_by_technology(self) -> pd.DataFrame:
        """RateOfUseByTechnology

        Notes
        -----
        From the formulation::

            r~REGION, l~TIMESLICE, t~TECHNOLOGY, f~FUEL, y~YEAR,
            sum{m in MODE_OF_OPERATION: InputActivityRatio[r,t,f,m,y]<>0}
                RateOfActivity[r,l,t,m,y] * InputActivityRatio[r,t,f,m,y]~VALUE;
        """
        try:
            rate_of_use_by_technology_by_mode = self[
                "RateOfUseByTechnologyByMode"
            ].copy(deep=True)
        except KeyError as ex:
            raise KeyError(self._msg("RateOfUseByTechnology", str(ex)))

        data = rate_of_use_by_technology_by_mode
        if not data.empty:
            data = data.groupby(
                by=["REGION", "TIMESLICE", "TECHNOLOGY", "FUEL", "YEAR"]
            ).sum()
        return data[(data != 0).all(1)]

    def rate_of_use_by_technology_by_mode(self) -> pd.DataFrame:
        """RateOfUseByTechnologyByMode

        Notes
        -----
        From the formulation::

            r~REGION, l~TIMESLICE, t~TECHNOLOGY, m~MODE_OF_OPERATION, f~FUEL, y~YEAR,
            RateOfActivity[r,l,t,m,y] * InputActivityRatio[r,t,f,m,y]~VALUE;
        """
        try:
            input_activity_ratio = self["InputActivityRatio"]
            rate_of_activity = self["RateOfActivity"]
        except KeyError as ex:
            raise KeyError(self._msg("RateOfUseByTechnology", str(ex)))

        data = input_activity_ratio.mul(rate_of_activity, fill_value=0.0)

        return data[(data != 0).all(1)]

    def total_annual_tech_activity_mode(self) -> pd.DataFrame:
        """TotalAnnualTechnologyActivityByMode

        Notes
        -----
        From the formulation::

            r~REGION, t~TECHNOLOGY, m~MODE_OF_OPERATION, y~YEAR,
            sum{l in TIMESLICE}
                RateOfActivity[r,l,t,m,y] * YearSplit[l,y]~VALUE;
        """
        try:
            rate_of_activity = self["RateOfActivity"]
            year_split = self["YearSplit"]
        except KeyError as ex:
            raise KeyError(self._msg("TotalAnnualTechnologyActivityByMode", str(ex)))

        data = rate_of_activity.mul(year_split, fill_value=0.0)
        return data[(data != 0).all(1)]

    def total_capacity_annual(self) -> pd.DataFrame:
        """Compute TotalCapacityAnnual result

        Notes
        -----
        From the formulation::

            r~REGION, t~TECHNOLOGY, y~YEAR,
            ResidualCapacity[r,t,y] +
            (sum{yy in YEAR: y-yy < OperationalLife[r,t] && y-yy>=0}
                NewCapacity[r,t,yy])~VALUE;
        """
        try:
            residual_capacity = self["ResidualCapacity"]
            acc_new_capacity = self["AccumulatedNewCapacity"]
        except KeyError as ex:
            raise KeyError(self._msg("TotalCapacityAnnual", str(ex)))

        data = residual_capacity.add(acc_new_capacity, fill_value=0.0)
        return data[(data != 0).all(1)]

    def total_discounted_cost(self) -> pd.DataFrame:
        """TotalDiscountedCost

        Notes
        -----
        From the formulation::

            r~REGION, y~YEAR,
            sum{t in TECHNOLOGY}
            ((((    (sum{yy in YEAR: y-yy
                         < OperationalLife[r,t]
                         && y-yy>=0}
                        NewCapacity[r,t,yy])
                        + ResidualCapacity[r,t,y])
                    * FixedCost[r,t,y]
                    + sum{m in MODE_OF_OPERATION, l in TIMESLICE}
                        RateOfActivity[r,l,t,m,y] * YearSplit[l,y]
                        * VariableCost[r,t,m,y])
                / ((1+DiscountRate[r,t])^(y-min{yy in YEAR} min(yy)+0.5))
                + CapitalCost[r,t,y] * NewCapacity[r,t,y]
                / ((1+DiscountRate[r,t])^(y-min{yy in YEAR} min(yy)))
                + DiscountedTechnologyEmissionsPenalty[r,t,y]
                - DiscountedSalvageValue[r,t,y]
            )    )
            + sum{r in REGION, s in STORAGE, y in YEAR}
                (CapitalCostStorage[r,s,y] * NewStorageCapacity[r,s,y]
                / ((1+DiscountRate[r,t])^(y-min{yy in YEAR} min(yy)))
                    - SalvageValueStorage[r,s,y]
                    / ((1+DiscountRate[r,t])^(max{yy in YEAR}
                        max(yy)-min{yy in YEAR} min(yy)+1))
                )~VALUE;
        """
        try:
            discount_rate = self["DiscountRate"]
            year_df = self["YEAR"].copy(deep=True)
            region_df = self["REGION"].copy(deep=True)

            years = year_df["VALUE"].tolist()
            regions = region_df["VALUE"].tolist()

            annual_fixed_operating_cost = self["AnnualFixedOperatingCost"]
            annual_variable_operating_cost = self["AnnualVariableOperatingCost"]
            capital_investment = self["CapitalInvestment"]

            technologies = self.get_unique_values_from_index(
                [
                    annual_fixed_operating_cost,
                    annual_variable_operating_cost,
                    capital_investment,
                ],
                "TECHNOLOGY",
            )

            discounted_emissions_penalty = self["DiscountedTechnologyEmissionsPenalty"]
            discounted_salvage_value = self["DiscountedSalvageValue"]

            # capital_cost_storage = self["CapitalCostStorage"]
        except KeyError as ex:
            raise KeyError(self._msg("TotalDiscountedCost", str(ex)))

        crf_op = capital_recovery_factor(
            regions, technologies, years, discount_rate, 0.5
        )
        crf_cap = capital_recovery_factor(
            regions, technologies, years, discount_rate, 0.0
        )

        undiscounted_operational_costs = annual_fixed_operating_cost.add(
            annual_variable_operating_cost, fill_value=0.0
        )
        discounted_operational_costs = undiscounted_operational_costs.div(
            crf_op, fill_value=0.0
        )
        discounted_capital_costs = capital_investment.div(crf_cap, fill_value=0.0)
        discounted_total_costs = discounted_operational_costs.add(
            discounted_capital_costs, fill_value=0.0
        )
        discounted_total_costs = discounted_total_costs.add(
            discounted_emissions_penalty, fill_value=0.0
        )
        discounted_total_costs = discounted_total_costs.sub(
            discounted_salvage_value, fill_value=0.0
        )

        # try:
        # new_storage_capacity = self["NewStorageCapacity"]
        # storage_investment = capital_investment.mul(
        #     new_storage_capacity, fill_value=0.0
        # )
        # except KeyError:
        #     LOGGER.info("Cannot find NewStorageCapacity, assuming empty")
        # storage_investment = pd.DataFrame()

        # try:
        # salvage_value_storage = self["SalvageValueStorage"]
        # except KeyError:
        #     LOGGER.info("Cannot find SalvageValueStorage, assuming empty")
        # salvage_value_storage = pd.DataFrame()

        data = discounted_total_costs

        if not data.empty:
            data = data.groupby(by=["REGION", "YEAR"]).sum()

        return data[(data != 0).all(1)].dropna()

    def get_unique_values_from_index(self, dataframes: List, name: str) -> List:
        """Utility function to extract list of unique values

        Extract unique values from the same index of the passed dataframes
        """
        elements = []  # type: List
        for df in dataframes:
            df = df.reset_index()
            if name in df.columns:
                elements += list(df[name].unique())
        return list(set(elements))

    def total_technology_annual_activity(self) -> pd.DataFrame:
        """TotalTechnologyAnnualActivity

        Notes
        -----
        From the formulation::

            ResultsPath & "/TotalTechnologyAnnualActivity.csv":
            r~REGION, t~TECHNOLOGY, y~YEAR,
            sum{l in TIMESLICE, m in MODE_OF_OPERATION}
                RateOfActivity[r,l,t,m,y] * YearSplit[l,y]~VALUE;
        """
        try:
            data = self["TotalAnnualTechnologyActivityByMode"].copy(deep=True)
        except KeyError as ex:
            raise KeyError(self._msg("TotalTechnologyAnnualActivity", str(ex)))

        if not data.empty:
            data = data.groupby(["REGION", "TECHNOLOGY", "YEAR"]).sum()

        return data[(data != 0).all(1)]

    def total_tech_model_period_activity(self) -> pd.DataFrame:
        """TotalTechnologyModelPeriodActivity

        Notes
        -----
        From the formulation::

            ResultsPath & "/TotalTechnologyModelPeriodActivity.csv":
            r~REGION, t~TECHNOLOGY,
            sum{l in TIMESLICE, m in MODE_OF_OPERATION, y in YEAR}
                RateOfActivity[r,l,t,m,y]*YearSplit[l,y]~VALUE;
        """
        try:
            data = self["TotalTechnologyAnnualActivity"].copy(deep=True)
        except KeyError as ex:
            raise KeyError(self._msg("TotalTechnologyModelPeriodActivity", str(ex)))

        if not data.empty:
            data = data.groupby(["REGION", "TECHNOLOGY"]).sum()

        return data[(data != 0).all(1)]

    def use_by_technology(self) -> pd.DataFrame:
        """UseByTechnology

        Notes
        -----
        From the formulation::

            r~REGION, l~TIMESLICE, t~TECHNOLOGY, f~FUEL, y~YEAR,
            sum{m in MODE_OF_OPERATION}
                RateOfActivity[r,l,t,m,y]
                * InputActivityRatio[r,t,f,m,y]
                * YearSplit[l,y]~VALUE;

        """
        try:
            rate_of_use = self["RateOfUseByTechnologyByMode"]
            year_split = self["YearSplit"]
        except KeyError as ex:
            raise KeyError(self._msg("UseByTechnology", str(ex)))

        data = rate_of_use.mul(year_split, fill_value=0.0)

        if not data.empty:
            data = data.groupby(
                ["REGION", "TIMESLICE", "TECHNOLOGY", "FUEL", "YEAR"]
            ).sum()

        return data[(data != 0).all(1)]


def capital_recovery_factor(
    regions: List,
    technologies: List,
    years: List,
    discount_rate: pd.DataFrame,
    adj: float = 0.0,
) -> pd.DataFrame:
    """Calculates the capital recovery factor

    Arguments
    ---------
    regions: list
    years: list
    discount_rate: pd.DataFrame
    adj: float, default=0.0
        Adjust to beginning of the year (default), mid year (0.5) or end year (1.0)
    """
    if regions and technologies and years:
        index = pd.MultiIndex.from_product(
            [regions, technologies, years], names=["REGION", "TECHNOLOGY", "YEAR"]
        )
        crf = discount_rate.reindex(index)
        crf = crf.reset_index(level="YEAR")
        crf["NUM"] = crf["YEAR"] - crf["YEAR"].min()
        crf["Rate"] = 1 + discount_rate
        crf["VALUE"] = crf["Rate"].pow(crf["NUM"] + adj)
        return crf.reset_index()[["REGION", "TECHNOLOGY", "YEAR", "VALUE"]].set_index(
            ["REGION", "TECHNOLOGY", "YEAR"]
        )
    else:
        return pd.DataFrame(
            [], columns=["REGION", "TECHNOLOGY", "YEAR", "VALUE"]
        ).set_index(["REGION", "TECHNOLOGY", "YEAR"])



"""Ensures that technology and fuel names match the convention

For example, to validate the following list of names, you would use the
config shown below::

    theseUNIQUE_ENTRY1
    are__UNIQUE_ENTRY2
    all__UNIQUE_ENTRY1
    validUNIQUE_ENTRY2
    entryUNIQUE_ENTRY1
    in__UNIQUE_ENTRY2
    a____UNIQUE_ENTRY1
    list_UNIQUE_ENTRY2

Create a yaml validation config with the following format::

    codes:
      some_valid_codes:
        UNIQUE_ENTRY1: Description of unique entry 1
        UNIQUE_ENTRY2: Description of unique entry 2
    schema:
      schema_name:
      - name: first_entry_in_schema
        valid: ['these', 'are__', 'all__', 'valid', 'entry', 'in__', 'a____', 'list_']
        position: (1, 5) # a tuple representing the start and end position
      - name: second_entry_in_schema
        valid: some_valid_codes  # references an entry in the codes section of the
        config
        position: (6, 19) # a tuple representing the start and end position

"""

import logging
import re
from collections import defaultdict
from typing import Dict, List, Sequence

import networkx.algorithms.isolate as isolate

from otoole.utils import read_datapackage, read_packaged_file
from otoole.visualise.res import create_graph

logger = logging.getLogger(__name__)


def read_validation_config():
    return read_packaged_file("validate.yaml", "otoole")


def check_for_duplicates(codes: Sequence) -> bool:
    duplicate_values = len(codes) != len(set(codes))
    return duplicate_values


def create_schema(config: Dict[str, Dict] = None) -> Dict:
    """Populate the dict of schema with codes from the validation config

    Arguments
    ---------
    config : dict, default=None
        A configuration dictionary containing ``codes`` and ``schema`` keys
    """
    if config is None:
        config = read_validation_config()

    for resource_name, resource_schemas in config["schema"].items():
        logger.debug("%s", resource_name)
        for schema in resource_schemas:

            for items in schema["items"]:  # typing: List

                if isinstance(items["valid"], str):
                    items["valid"] = list(
                        config["codes"][items["valid"]].keys()
                    )  # typing: List
                    logger.debug("create_schema: %s", items["valid"])
                elif isinstance(items["valid"], list):
                    pass
                else:
                    raise ValueError("Entry {} is not correct".format(schema["name"]))

                if check_for_duplicates(items["valid"]):
                    raise ValueError(
                        "There are duplicate values in codes for {}", schema["name"]
                    )

    return config["schema"]


def compose_expression(schema: List) -> str:
    """Generates a regular expression from a schema

    Returns
    -------
    str
    """
    expression = "^"
    for x in schema:
        logger.debug("compose_expression: %s", x["valid"])
        valid_entries = "|".join(x["valid"])
        expression += "({})".format(valid_entries)
    return expression


def compose_multi_expression(resource: List) -> str:
    """Concatenates multiple expressions using an OR operator

    Use to validate elements using an OR operation e.g. the elements
    must match this expression OR the expression
    """
    expressions = []
    for schemas in resource:
        expressions.append(compose_expression(schemas["items"]))
    return "|".join(expressions)


def validate(expression: str, name: str) -> bool:
    """Determine if ``name`` matches the ``expression``

    Arguments
    ---------
    expression : str
    name : str

    Returns
    -------
    bool
    """
    logger.debug("Running validation for %s", name)

    valid = False

    pattern = re.compile(expression)

    if pattern.fullmatch(name):
        valid = True
    else:
        valid = False
    return valid


def validate_resource(package, resource: str, schemas: List[Dict]):
    """

    Arguments
    ---------
    package
    resource: str
    schemas : List[Dict]
        The schema from which to create a validation expression
    """

    print(
        "Validating {} with {}\n".format(
            resource, ", ".join([x["name"] for x in schemas])
        )
    )

    logger.debug(schemas)

    expression = compose_multi_expression(schemas)
    resources = package.get_resource(resource).read(keyed=True)

    valid_names = []
    invalid_names = []

    for row in resources:
        name = row["VALUE"]
        valid = validate(expression, row["VALUE"])
        if valid:
            valid_names.append(name)
        else:
            invalid_names.append(name)

    if invalid_names:
        msg = "{} invalid names:\n{}\n"
        print(msg.format(len(invalid_names), ", ".join(invalid_names)))
    if valid_names:
        msg = "{} valid names:\n{}\n"
        print(msg.format(len(valid_names), ", ".join(valid_names)))


def identify_orphaned_fuels_techs(package) -> Dict[str, str]:
    """Returns a list of fuels and technologies which are unconnected

    Returns
    -------
    dict

    """
    graph = create_graph(package)

    number_of_isolates = isolate.number_of_isolates(graph)
    logger.debug("There are {} isolated nodes in the graph".format(number_of_isolates))

    isolated_nodes: Dict = defaultdict(list)

    for node_name in list(isolate.isolates(graph)):
        node_data = graph.nodes[node_name]
        isolated_nodes[node_data["type"]].append(node_name)

    return isolated_nodes


def main(file_format: str, filepath: str, config=None):

    print("\n***Beginning validation***\n")
    if file_format == "datapackage":
        package = read_datapackage(filepath)
    elif file_format == "sql":
        package = read_datapackage(filepath, sql=True)
    else:
        package = None

    schema = create_schema(config)

    if package:

        for resource, schemas in schema.items():
            validate_resource(package, resource, schemas)

        print("\n***Checking graph structure***")
        isolated_nodes = identify_orphaned_fuels_techs(package)

        msg = ""
        for node_type, node_names in isolated_nodes.items():
            msg += "\n{} '{}' nodes are isolated:\n     {}\n".format(
                len(node_names), node_type, ", ".join(node_names)
            )
        print(msg)

    else:
        raise ValueError("Unable to load the datapackage")



# -*- coding: utf-8 -*-
from pkg_resources import get_distribution, DistributionNotFound

from otoole.input import Context
from otoole.read_strategies import (
    ReadCsv,
    ReadDatafile,
    ReadDatapackage,
    ReadExcel,
    ReadMemory,
)
from otoole.write_strategies import (
    WriteCsv,
    WriteDatafile,
    WriteDatapackage,
    WriteExcel,
)
from otoole.results.results import ReadCbc, ReadCplex

try:
    # Change here if project is renamed and does not equal the package name
    dist_name = __name__
    __version__ = get_distribution(dist_name).version
except DistributionNotFound:
    __version__ = "unknown"
finally:
    del get_distribution, DistributionNotFound


__all__ = [
    "Context",
    "ReadCbc",
    "ReadCsv",
    "ReadCplex",
    "ReadDatafile",
    "ReadDatapackage",
    "ReadExcel",
    "ReadMemory",
    "WriteCsv",
    "WriteDatafile",
    "WriteDatapackage",
    "WriteExcel",
]
